from datetime import date
from database import db, ITEMS_PER_PAGE, cache
from sqlalchemy.orm import subqueryload
from sqlalchemy import not_, and_


class BlueprintEntity(db.Model):
    __tablename__ = 'all_entity'

    id = db.Column(db.Integer, primary_key=True)
    processed_update = db.Column(db.Boolean, nullable=False)
    processed_full_update = db.Column(db.Boolean, nullable=False)

    @classmethod
    def remove_all(cls):
        num_rows_deleted = db.session.query(cls).delete()
        db.session.commit()
        return num_rows_deleted

    @classmethod
    def delete_all(cls):
        db.session.query(cls).delete()
        db.session.commit()
        return None

    @classmethod
    @cache.memoize(50)
    def get_all_as_page(cls, page: int):
        return db.session.query(cls)\
            .paginate(page, per_page=ITEMS_PER_PAGE)

    @classmethod
    @cache.memoize(50)
    def get_all(cls, page: int):
        return db.session.query(cls).paginate(page, per_page=ITEMS_PER_PAGE)

    @classmethod
    @cache.cached(timeout=50)
    def find_all(cls):
        return db.session.query(cls).all()

    @classmethod
    def find_all_as_dict(cls):
        pass

    @classmethod
    @cache.memoize(50)
    def get_by_id(cls, other_id):
        return db.session.query(cls).filter(cls.id == other_id).one()

    @classmethod
    @cache.memoize(50)
    def find_by_id(cls, other_id):
        return db.session.query(cls).filter(cls.id == other_id).one_or_none()

    @classmethod
    @cache.cached(timeout=50)
    def set_all_processed_full_update(cls):
        for o in cls.find_by_not_processed_full_update():
            o.set_processed_full_update()
        db.session.commit()
        return None

    def set_processed_update(self):
        self.processed_update = True
        return self

    def set_processed_full_update(self):
        self.processed_full_update = True
        return self

    def unset_processed_update(self):
        self.processed_update = False
        return self

    def unset_processed_full_update(self):
        self.processed_full_update = False
        return self

    @classmethod
    @cache.cached(timeout=50)
    def find_by_not_processed_update(cls):
        return db.session.query(cls).filter(not_(cls.processed_update)).all()

    @classmethod
    @cache.cached(timeout=50)
    def find_by_not_processed_full_update(cls):
        return db.session.query(cls).filter(not_(cls.processed_full_update)).all()


class BlueprintDateReported(BlueprintEntity):
    __tablename__ = 'all_date_reported'
    __mapper_args__ = {'concrete': True}
    __table_args__ = (
        db.UniqueConstraint(
            'date_reported_import_str',
            'datum',
            'year_day_of_year',
            name="uix_all_date_reported"
        ),
    )

    def __str__(self):
        return self.datum.isoformat()

    id = db.Column(db.Integer, primary_key=True)
    processed_update = db.Column(db.Boolean, nullable=False)
    processed_full_update = db.Column(db.Boolean, nullable=False)
    #
    date_reported_import_str = db.Column(db.String(255), nullable=False, unique=True)
    datum = db.Column(db.Date, nullable=False, unique=True)
    #
    year_day_of_year = db.Column(db.String(255), nullable=False, unique=True)
    #
    year_month = db.Column(db.String(255), nullable=False)
    year_week = db.Column(db.String(255), nullable=False)
    #
    year = db.Column(db.Integer, nullable=False)
    month = db.Column(db.Integer, nullable=False)
    day_of_month = db.Column(db.Integer, nullable=False)
    day_of_week = db.Column(db.Integer, nullable=False)
    day_of_year = db.Column(db.Integer, nullable=True)
    week_of_year = db.Column(db.Integer, nullable=False)

    def get_name_for_weekday(self):
        return self.get_names_for_weekday()[self.day_of_week]

    @classmethod
    @cache.cached(timeout=50)
    def get_names_for_weekday(cls):
        return {1: "Montag", 2: "Dienstag", 3: "Mittwoch", 4: "Donnerstag",
                5: "Freitag", 6: "Samstag", 7: "Sonntag"}

    def get_name_for_month(self):
        return self.get_names_for_months()[self.month]

    @classmethod
    @cache.cached(timeout=50)
    def get_names_for_months(cls):
        return {1: "Januar", 2: "Februar", 3: "März", 4: "April", 5: "Mai", 6: "Juni",
                7: "Juli", 8: "August", 9: "September", 10: "Oktober", 11: "November", 12: "Dezember"}

    @classmethod
    @cache.memoize(50)
    def get_by_datum(cls, datum: date):
        return db.session.query(cls).filter(cls.datum == datum).one()

    @classmethod
    def get_by_date_reported(cls, date_reported_import_str: str):
        return db.session.query(cls)\
            .filter(cls.date_reported_import_str == date_reported_import_str)\
            .one()

    @classmethod
    @cache.memoize(50)
    def find_by_date_reported(cls, date_reported_import_str: str):
        return db.session.query(cls)\
            .filter(cls.date_reported_import_str == date_reported_import_str)\
            .one_or_none()

    @classmethod
    @cache.memoize(50)
    def find_by_year_week(cls, year_week: str):
        return db.session.query(cls)\
            .filter(cls.year_week == year_week)\
            .all()

    @classmethod
    @cache.cached(timeout=50)
    def get_joungest_datum(cls):
        return db.session.query(cls)\
            .order_by(cls.datum.desc())\
            .first()

    @classmethod
    @cache.cached(timeout=50)
    def set_all_processed_update(cls):
        for o in cls.find_by_not_processed_update():
            o.set_processed_update()
        db.session.commit()
        return None

    @classmethod
    def find_all(cls):
        return db.session.query(cls).order_by(cls.datum.desc()).all()

    @classmethod
    def find_all_as_dict(cls):
        dates_reported = {}
        for my_date_reported in cls.find_all():
            dates_reported[my_date_reported.date_reported_import_str] = my_date_reported
        return dates_reported


class BlueprintLocationGroup(BlueprintEntity):
    __tablename__ = 'all_location_group'
    __mapper_args__ = {'concrete': True}
    __table_args__ = (
        db.UniqueConstraint('location_group', name='uix_all_location_group'),
    )

    def __str__(self):
        result = " " + self.location_group + " "
        return result

    id = db.Column(db.Integer, primary_key=True)
    processed_update = db.Column(db.Boolean, nullable=False)
    processed_full_update = db.Column(db.Boolean, nullable=False)
    location_group = db.Column(db.String(255), nullable=False, unique=True)

    @classmethod
    @cache.cached(timeout=50)
    def get_last(cls):
        return db.session.query(cls).order_by(cls.location_group).all().pop()

    @classmethod
    @cache.memoize(50)
    def get_by_location_group(cls, location_group: str):
        return db.session.query(cls)\
            .filter(cls.location_group == location_group)\
            .one()

    @classmethod
    # @cache.memoize(50)
    def find_by_location_group(cls, location_group: str):
        return db.session.query(cls) \
            .filter(cls.location_group == location_group) \
            .one_or_none()

    @classmethod
    @cache.memoize(50)
    def get_all_as_page(cls, page: int):
        return db.session.query(cls)\
            .order_by(cls.location_group)\
            .paginate(page, per_page=ITEMS_PER_PAGE)

    @classmethod
    @cache.cached(timeout=50)
    def find_all(cls):
        return db.session.query(cls).order_by(cls.location_group).all()

    @classmethod
    @cache.cached(timeout=50)
    def find_all_as_dict(cls):
        dates_reported = {}
        for my_location_group in cls.find_all():
            dates_reported[my_location_group.location_group] = my_location_group
        return dates_reported


class BlueprintLocation(BlueprintEntity):
    __tablename__ = 'all_location'
    __mapper_args__ = {'concrete': True}
    __table_args__ = (
        db.UniqueConstraint('location', name='uix_all_location'),
    )

    def __str__(self):
        return self.location_group.__str__() + " : " + self.location_code + " | " + self.location

    id = db.Column(db.Integer, primary_key=True)
    processed_update = db.Column(db.Boolean, nullable=False)
    processed_full_update = db.Column(db.Boolean, nullable=False)
    location_code = db.Column(db.String(255), unique=True, nullable=True)
    location = db.Column(db.String(255), nullable=False, unique=True)
    location_group_id = db.Column(db.Integer, db.ForeignKey('all_location_group.id'), nullable=False)
    location_group = db.relationship(
        'BlueprintLocationGroup',
        lazy='joined',
        cascade='save-update',
        order_by='BlueprintLocationGroup.location_group')

    @classmethod
    @cache.memoize(50)
    def find_by_location_code(cls, location_code: str):
        return db.session.query(cls).filter(cls.location_code == location_code)\
            .one_or_none()

    @classmethod
    @cache.memoize(50)
    def get_by_location_code(cls, location_code: str):
        return db.session.query(cls).filter(cls.location_code == location_code)\
            .one()

    @classmethod
    @cache.memoize(50)
    def find_by_location(cls, location: str):
        return db.session.query(cls).filter(cls.location == location)\
            .one_or_none()

    @classmethod
    @cache.memoize(50)
    def get_by_location(cls, location: str):
        return db.session.query(cls).filter(cls.location == location)\
            .one()

    @classmethod
    @cache.memoize(50)
    def find_by_location_group(cls, location_group: BlueprintLocationGroup):
        return db.session.query(cls).filter(cls.location_group == location_group)\
            .order_by(cls.location)\
            .all()

    @classmethod
    @cache.memoize(50)
    def get_by_location_group(cls, location_group: BlueprintLocationGroup, page: int):
        return db.session.query(cls).filter(cls.location_group == location_group)\
            .order_by(cls.location)\
            .paginate(page, per_page=ITEMS_PER_PAGE)

    @classmethod
    @cache.memoize(50)
    def find_by_location_code_and_location_and_location_group(cls, location_code: str, location: str,
                                                              location_group: BlueprintLocationGroup):
        return db.session.query(cls).filter(
            and_(
                cls.location_code == location_code,
                cls.location == location,
                cls.location_group_id == location_group.id
            )
        ).one_or_none()

    @classmethod
    @cache.memoize(50)
    def get_by_location_code_and_location_and_location_group(cls, location_code: str, location: str,
                                                             location_group: BlueprintLocationGroup):
        return db.session.query(cls).filter(
            and_(
                cls.location_code == location_code,
                cls.location == location,
                cls.location_group_id == location_group.id
            )
        ).one()

    @classmethod
    @cache.memoize(50)
    def find_by_location_code_and_location(cls, location_code: str, location: str):
        return db.session.query(cls)\
            .filter(and_(cls.location_code == location_code, cls.location == location))\
            .order_by(cls.location)\
            .one_or_none()

    @classmethod
    @cache.memoize(50)
    def get_by_location_code_and_location(cls, location_code: str, location: str):
        return db.session.query(cls).filter(and_(cls.location_code == location_code, cls.location == location)) \
            .one()

    @classmethod
    @cache.cached(timeout=50)
    def find_all(cls):
        return db.session.query(cls).order_by(cls.location).all()

    @classmethod
    @cache.cached(timeout=50)
    def find_all_as_dict(cls):
        dates_reported = {}
        for my_location in cls.find_all():
            dates_reported[my_location.location] = my_location
        return dates_reported

    @classmethod
    @cache.memoize(50)
    def get_all_as_page(cls, page: int):
        return db.session.query(cls)\
            .order_by(cls.location)\
            .paginate(page, per_page=ITEMS_PER_PAGE)


class BlueprintFactTableTimeSeries(BlueprintEntity):
    __tablename__ = 'all_data_timeline'
    __mapper_args__ = {'concrete': True}
    __table_args__ = (
        db.UniqueConstraint('date_reported_id', name='uix_all_data_timeline'),
    )

    def __str__(self):
        return self.date_reported.__str__()

    id = db.Column(db.Integer, primary_key=True)
    processed_update = db.Column(db.Boolean, nullable=False)
    processed_full_update = db.Column(db.Boolean, nullable=False)
    date_reported_id = db.Column(db.Integer, db.ForeignKey('all_date_reported.id'), nullable=False)
    date_reported = db.relationship(
        'BlueprintDateReported',
        lazy='joined',
        cascade='save-update',
        order_by='desc(BlueprintDateReported.datum)')

    @classmethod
    @cache.cached(timeout=50)
    def get_date_reported_list(cls):
        date_reported_list = []
        for data in db.session.query(cls).options(subqueryload("date_reported").load_only("datum")):
            datum = data.date_reported.datum.isoformat()
            if not datum in date_reported_list:
                date_reported_list.append(datum)
        date_reported_list.sort()
        return date_reported_list

    @classmethod
    @cache.cached(timeout=50)
    def get_joungest_date_reported(cls):
        data = cls.get_date_reported_list()
        if len(data) > 0:
            return data.pop()
        else:
            return None

    @classmethod
    def find_by_date_reported(cls, date_reported: BlueprintDateReported):
        pass

    @classmethod
    def get_by_date_reported(cls, date_reported: BlueprintDateReported, page: int):
        pass

    @classmethod
    def delete_data_for_one_day(cls, date_reported: BlueprintDateReported):
        pass


class BlueprintFactTable(BlueprintFactTableTimeSeries):
    __tablename__ = 'all_data'
    __mapper_args__ = {'concrete': True}
    __table_args__ = (
        db.UniqueConstraint('location_id', 'date_reported_id', name='uix_all_data'),
    )

    def __str__(self):
        return self.date_reported.__str__() + " " + self.location.__str__()

    id = db.Column(db.Integer, primary_key=True)
    processed_update = db.Column(db.Boolean, nullable=False)
    processed_full_update = db.Column(db.Boolean, nullable=False)
    date_reported_id = db.Column(db.Integer, db.ForeignKey('all_date_reported.id'), nullable=False)
    date_reported = db.relationship(
        'BlueprintDateReported',
        lazy='joined',
        cascade='save-update',
        order_by='desc(BlueprintDateReported.datum)')
    location_id = db.Column(db.Integer, db.ForeignKey('all_location.id'), nullable=False)
    location = db.relationship(
        'BlueprintLocation',
        lazy='joined',
        cascade='save-update',
        order_by='asc(BlueprintLocation.location)')

    @classmethod
    def get_by_location(cls, location: BlueprintLocation, page: int):
        pass

    @classmethod
    def find_by_location(cls, location: BlueprintLocation):
        pass

    @classmethod
    def find_by_date_reported_and_location(cls, date_reported: BlueprintDateReported, location: BlueprintLocation):
        pass

    @classmethod
    def get_by_date_reported_and_location(cls, date_reported: BlueprintDateReported, location: BlueprintLocation, page: int):
        pass


class AllImport(BlueprintEntity):
    __tablename__ = 'all_import'
    __mapper_args__ = {'concrete': True}

    def __str__(self):
        return " [ " + self.datum.isoformat() + " ] " \
               + "updated: " + str(self.processed_update) \
               + " | full_updated: " + str(self.processed_full_update)

    id = db.Column(db.Integer, primary_key=True)
    processed_update = db.Column(db.Boolean, nullable=False)
    processed_full_update = db.Column(db.Boolean, nullable=False)
    #
    date_reported_import_str = db.Column(db.String(255), nullable=False)
    datum = db.Column(db.Date, nullable=False)

    @classmethod
    def find_by_datum(cls, datum: date):
        return db.session.query(cls) \
            .filter(cls.datum == datum) \
            .all()

    @classmethod
    @cache.cached(timeout=50)
    def get_datum_list(cls):
        return db.session.query(cls.datum) \
            .order_by(cls.datum.desc()) \
            .distinct() \
            .all()


class AllFlat(AllImport):
    __tablename__ = 'all_import_flat'
    __mapper_args__ = {'concrete': True}

    def __str__(self):
        return " [ " + self.datum.isoformat() + " ] " \
               + self.location_group + " : " \
               + self.location_code + " | " \
               + self.location\
               + " | updated: " + str(self.processed_update) \
               + " | full_updated: " + str(self.processed_full_update)

    id = db.Column(db.Integer, primary_key=True)
    processed_update = db.Column(db.Boolean, nullable=False)
    processed_full_update = db.Column(db.Boolean, nullable=False)
    #
    date_reported_import_str = db.Column(db.String(255), nullable=False)
    datum = db.Column(db.Date, nullable=False)
    location = db.Column(db.String(255), nullable=False)
    location_group = db.Column(db.String(255), nullable=False)
    location_code = db.Column(db.String(255), nullable=False)
    #
    year = db.Column(db.Integer, nullable=False)
    year_month = db.Column(db.String(255), nullable=False)
    year_week = db.Column(db.String(255), nullable=False)
    year_day_of_year = db.Column(db.String(255), nullable=False)
    #
    month = db.Column(db.Integer, nullable=False)
    day_of_month = db.Column(db.Integer, nullable=False)
    day_of_week = db.Column(db.Integer, nullable=False)
    week_of_year = db.Column(db.Integer, nullable=False)
    day_of_year = db.Column(db.Integer, nullable=False)
    #

