from sqlalchemy import and_
from sqlalchemy.orm import joinedload
from database import db, ITEMS_PER_PAGE, cache
from flask_covid19.blueprints.app_all.all_model import BlueprintDateReported, BlueprintLocationGroup
from flask_covid19.blueprints.app_all.all_model import BlueprintLocation, BlueprintFactTable


class EcdcDateReported(BlueprintDateReported):
    __tablename__ = 'ecdc_date_reported'
    __mapper_args__ = {'concrete': True}
    __table_args__ = (
        db.UniqueConstraint(
            'date_reported_import_str',
            'datum',
            'year_day_of_year',
            name="uix_ecdc_date_reported"
        ),
    )

    def __repr__(self):
        return "%s(%s, %s, %s)" % (self.__class__.__name__,
                                   self.date_reported_import_str,
                                   self.datum.isoformat(),
                                   self.year_day_of_year)

    id = db.Column(db.Integer, primary_key=True)
    processed_update = db.Column(db.Boolean, nullable=False)
    processed_full_update = db.Column(db.Boolean, nullable=False)
    #
    date_reported_import_str = db.Column(db.String(255), nullable=False, unique=True)
    datum = db.Column(db.Date, nullable=False, unique=True)
    #
    year_day_of_year = db.Column(db.String(255), nullable=False, unique=True)
    #
    year_month = db.Column(db.String(255), nullable=False)
    year_week = db.Column(db.String(255), nullable=False)
    #
    year = db.Column(db.Integer, nullable=False)
    month = db.Column(db.Integer, nullable=False)
    day_of_month = db.Column(db.Integer, nullable=False)
    day_of_week = db.Column(db.Integer, nullable=False)
    day_of_year = db.Column(db.Integer, nullable=True)
    week_of_year = db.Column(db.Integer, nullable=False)

    @classmethod
    @cache.memoize(50)
    def get_all_as_page(cls, page: int):
        return db.session.query(cls)\
            .order_by(cls.datum.desc())\
            .paginate(page, per_page=ITEMS_PER_PAGE)


class EcdcContinent(BlueprintLocationGroup):
    __tablename__ = 'ecdc_country_continent'
    __mapper_args__ = {'concrete': True}
    __table_args__ = (
        db.UniqueConstraint('location_group', name="uix_ecdc_country_continent"),
    )

    def __repr__(self):
        return "%s(%s)" % (self.__class__.__name__, self.location_group)

    id = db.Column(db.Integer, primary_key=True)
    processed_update = db.Column(db.Boolean, nullable=False)
    processed_full_update = db.Column(db.Boolean, nullable=False)
    location_group = db.Column(db.String(255), nullable=False, unique=True)


class EcdcCountry(BlueprintLocation):
    __tablename__ = 'ecdc_country'
    __mapper_args__ = {'concrete': True}
    __table_args__ = (
        db.UniqueConstraint(
            'location',
            'geo_id',
            'location_code',
            name="uix_ecdc_country"
        ),
    )

    def __repr__(self):
        return "%s(%s, %s, %s; %s, %s)" % (self.__class__.__name__,
                                           self.location_group.__repr__(),
                                           self.location_code,
                                           self.location,
                                           self.pop_data_2019,
                                           self.geo_id)

    def __str__(self):
        return " " + self.location_group.location_group \
             + " : " + self.geo_id \
             + " | " + self.location_code \
             + " | " + self.location \
             + " | " + self.pop_data_2019 \


    id = db.Column(db.Integer, primary_key=True)
    processed_update = db.Column(db.Boolean, nullable=False)
    processed_full_update = db.Column(db.Boolean, nullable=False)
    #
    # country_territory_code = db.Column(db.String(255), nullable=False)
    location_code = db.Column(db.String(255), nullable=True)
    # countries_and_territories = db.Column(db.String(255), nullable=False)
    location = db.Column(db.String(255), nullable=False, index=True)
    location_group_id = db.Column(db.Integer, db.ForeignKey('ecdc_country_continent.id'), nullable=False)
    location_group = db.relationship(
        'EcdcContinent',
        lazy='subquery', cascade="all, delete",
        order_by='asc(EcdcContinent.location_group)'
    )
    pop_data_2019 = db.Column(db.String(255), nullable=False)
    geo_id = db.Column(db.String(255), nullable=False)

    @classmethod
    def get_by(cls, location: str = '', geo_id: str = '', location_code: str = ''):
        return db.session.query(cls).filter(and_(
            (cls.location == location),
            (cls.geo_id == geo_id),
            (cls.location_code == location_code)
        )).one()

    @classmethod
    @cache.memoize(50)
    def find_by(cls, location: str = '', geo_id: str = '', location_code: str = ''):
        return db.session.query(cls).filter(and_(
            (cls.location == location),
            (cls.geo_id == geo_id),
            (cls.location_code == location_code)
        )).one_or_none()

    @classmethod
    @cache.memoize(50)
    def find_by_location_group(cls, location_group: EcdcContinent):
        return db.session.query(cls)\
            .filter(cls.location_group_id == location_group.id)\
            .all()

    @classmethod
    @cache.memoize(50)
    def get_by_location_group(cls, location_group: EcdcContinent, page: int):
        return db.session.query(cls)\
            .filter(cls.location_group_id == location_group.id)\
            .paginate(page, per_page=ITEMS_PER_PAGE)

    @classmethod
    @cache.memoize(50)
    def find_germany(cls):
        return db.session.query(cls) \
            .filter(cls.location_code == 'DEU') \
            .one_or_none()


class EcdcData(BlueprintFactTable):
    __tablename__ = 'ecdc'
    __mapper_args__ = {'concrete': True}
    __table_args__ = (
        db.UniqueConstraint('date_reported_id', 'location_id', name="uix_ecdc"),
    )

    def __repr__(self):
        return "%s(%s %s)" % (self.__class__.__name__, self.date_reported.__repr__(), self.location.__repr__())

    id = db.Column(db.Integer, primary_key=True)
    processed_update = db.Column(db.Boolean, nullable=False)
    processed_full_update = db.Column(db.Boolean, nullable=False)
    #
    date_reported_id = db.Column(db.Integer, db.ForeignKey('ecdc_date_reported.id'), nullable=False)
    date_reported = db.relationship(
        'EcdcDateReported',
        lazy='joined',
        cascade='save-update',
        order_by='desc(EcdcDateReported.datum)')
    location_id = db.Column(db.Integer, db.ForeignKey('ecdc_country.id'), nullable=False)
    location = db.relationship(
        'EcdcCountry',
        lazy='joined',
        cascade='save-update',
        order_by='asc(EcdcCountry.location)')
    #
    deaths = db.Column(db.Integer, nullable=False)
    cases = db.Column(db.Integer, nullable=False)
    cumulative_number_for_14_days_of_covid19_cases_per_100000 = db.Column(db.Float, nullable=False)

    @classmethod
    def __query_by_date_reported(cls, date_reported: EcdcDateReported):
        return db.session.query(cls).filter(
            cls.date_reported_id == date_reported.id
        ).populate_existing().options(
            joinedload(cls.location).joinedload(EcdcDateReported.location_group),
            joinedload(cls.date_reported)
        )

    @classmethod
    def __query_by_location(cls, location: EcdcCountry):
        return db.session.query(cls).filter(
            cls.location_id == location.id
        ).populate_existing().options(
            joinedload(cls.location).joinedload(EcdcCountry.location_group),
            joinedload(cls.date_reported)
        )

    @classmethod
    def __query_by_date_reported_order_by_notification_rate(cls, date_reported: EcdcDateReported):
        return cls.__query_by_date_reported(date_reported) \
            .order_by(cls.cumulative_number_for_14_days_of_covid19_cases_per_100000.desc())

    @classmethod
    def __query_by_date_reported_order_by_deaths(cls, date_reported: EcdcDateReported):
        return cls.__query_by_date_reported(date_reported) \
            .order_by(cls.deaths.desc())

    @classmethod
    def __query_by_date_reported_order_by_cases(cls, date_reported: EcdcDateReported):
        return cls.__query_by_date_reported(date_reported) \
            .order_by(cls.cases.desc())

    @classmethod
    def __query_by_date_reported_and_location(cls, date_reported: EcdcDateReported, location: EcdcCountry):
        return db.session.query(cls) \
            .filter(and_((cls.location_id == location.id), (cls.date_reported_id == date_reported.id)))

    @classmethod
    @cache.memoize(50)
    def find_by_date_reported(cls, date_reported: EcdcDateReported):
        return cls.__query_by_date_reported(date_reported).all()

    @classmethod
    @cache.memoize(50)
    def get_by_date_reported(cls, date_reported: EcdcDateReported, page: int):
        return cls.__query_by_date_reported(date_reported)\
            .paginate(page, per_page=ITEMS_PER_PAGE)

    @classmethod
    @cache.memoize(50)
    def find_by_date_reported_order_by_notification_rate(cls, date_reported: EcdcDateReported):
        return cls.__query_by_date_reported_order_by_notification_rate(date_reported).all()

    @classmethod
    @cache.memoize(50)
    def get_by_date_reported_order_by_notification_rate(cls, date_reported: EcdcDateReported, page: int):
        return cls.__query_by_date_reported_order_by_notification_rate(date_reported)\
            .paginate(page, per_page=ITEMS_PER_PAGE)

    @classmethod
    @cache.memoize(50)
    def find_by_date_reported_order_by_deaths(cls, date_reported: EcdcDateReported):
        return cls.__query_by_date_reported_order_by_deaths_weekly(date_reported).all()

    @classmethod
    @cache.memoize(50)
    def get_by_date_reported_order_by_deaths(cls, date_reported: EcdcDateReported, page: int):
        return cls.__query_by_date_reported_order_by_deaths_weekly(date_reported)\
            .paginate(page, per_page=ITEMS_PER_PAGE)

    @classmethod
    @cache.memoize(50)
    def find_by_date_reported_order_by_cases(cls, date_reported: EcdcDateReported):
        return cls.__query_by_date_reported_order_by_cases_weekly(date_reported).all()

    @classmethod
    @cache.memoize(50)
    def get_by_date_reported_order_by_cases(cls, date_reported: EcdcDateReported, page: int):
        return cls.__query_by_date_reported_order_by_cases_weekly(date_reported)\
            .paginate(page, per_page=ITEMS_PER_PAGE)

    @classmethod
    @cache.memoize(50)
    def find_by_location(cls, location: EcdcCountry):
        return cls.__query_by_location(location).all()

    @classmethod
    @cache.memoize(50)
    def get_by_location(cls, location: EcdcCountry, page: int):
        return cls.__query_by_location(location).paginate(page, per_page=ITEMS_PER_PAGE)

    @classmethod
    @cache.memoize(50)
    def find_by_date_reported_and_location(cls, date_reported: EcdcDateReported, location: EcdcCountry):
        return cls.__query_by_date_reported_and_location(date_reported, location).one_or_none()

    @classmethod
    @cache.memoize(50)
    def get_by_date_reported_and_location(cls, date_reported: EcdcDateReported, location: EcdcCountry, page: int):
        return cls.__query_by_date_reported_and_location(date_reported, location).one()

    @classmethod
    def delete_data_for_one_day(cls, date_reported: EcdcDateReported):
        for data in cls.find_by_date_reported(date_reported):
            db.session.delete(data)
        db.session.delete(date_reported)
        db.session.commit()

